#ifndef MYBOT_CONTROLLER_H
#define MYBOT_CONTROLLER_H

/*
 * Include some necessary headers.
 */

/* Definition of the CCI_Controller class. */
#include <argos3/core/control_interface/ci_controller.h>
/* Definition of the differential steering actuator */
#include <argos3/plugins/robots/generic/control_interface/ci_differential_steering_actuator.h>

#include <argos3/plugins/robots/foot-bot/control_interface/ci_footbot_proximity_sensor.h>
#include <argos3/plugins/robots/generic/control_interface/ci_leds_actuator.h>
#include <argos3/plugins/robots/generic/control_interface/ci_colored_blob_omnidirectional_camera_sensor.h>
#include <argos3/plugins/robots/foot-bot/control_interface/ci_footbot_light_sensor.h>


#include "../../mybot.h"

using namespace argos;


class MybotController : public CCI_Controller {

public:
    virtual ~MybotController() = default;

    /*
     * This function initializes the controller.
     * The 't_node' variable points to the <parameters> section in the XML
     * file in the <controllers><footbot_diffusion_controller> section.
     */
    virtual void Init(TConfigurationNode& t_node);

    /*
     * This function is called once every time step.
     * The length of the time step is set in the XML file.
     */
    virtual void ControlStep() = 0;

    /*
     * This function resets the controller to its state right after the
     * Init().
     * It is called when you press the reset button in the GUI.
     * In this example controller there is no need for resetting anything,
     * so the function could have been omitted. It's here just for
     * completeness.
     */
    virtual void Reset() {}

    /*
     * Called to cleanup what done by Init() when the experiment finishes.
     * In this example controller there is no need for clean anything up,
     * so the function could have been omitted. It's here just for
     * completeness.
     */
    virtual void Destroy() {}

private:


protected:
    Mybot mybot;
};


#define MYBOT_CONTROLLER_HEADER(CLASS_NAME) \
    class CLASS_NAME : public MybotController{

#define MYBOT_CONTROLLER_FOOTER(CLASS_NAME, PROGRAM_NAME)                                         \
public:                                                                                             \
    ~CLASS_NAME() = default;                                                                        \
                                                                                                    \
    void Init(TConfigurationNode& t_node) override {                                                \
        MybotController::Init(t_node);                                                            \
        if ( ! mybot.isInitialized() ) {                                                          \
            throw std::runtime_error("mybot.ino.cpp: mybot wasn't initialized in controller");  \
        }                                                                                           \
        setup();                                                                                    \
    }                                                                                               \
                                                                                                    \
    void ControlStep() override {                                                                   \
        loop();                                                                                     \
    }                                                                                               \
                                                                                                    \
    void Reset() override {                                                                         \
        MybotController::Reset();                                                                 \
        setup();                                                                                    \
    }                                                                                               \
};                                                                                                  \
                                                                                                    \
/* \
 * This statement notifies ARGoS of the existence of the controller. \
 * It binds the class passed as first argument to the string passed as \
 * second argument. \
 * The string is then usable in the configuration file to refer to this \
 * controller. \
 * When ARGoS reads that string in the configuration file, it knows which \
 * controller class to instantiate. \
 * See also the configuration files for an example of how this is used. \
 */\
REGISTER_CONTROLLER(CLASS_NAME, PROGRAM_NAME)

#endif
